from structs import Operation, Predicate


class PrintOperation(Operation):
    """
    Simple operation to print whatever input is supplied.
    """
    def perform(self, obj):
        print obj


class TruePredicate(Predicate):
    """
    Simple predicate that always returns True.
    """
    def test(self, obj):
        return True


class NullOperation(Operation):
    """
    An empty operation that does nothing, it is simply used to 'pull' objects through the
    previous operations in the stream pipeline.
    """
    def perform(self, obj):
        pass



import abc


class Operation(object):
    """
    Defines an operation that can be applied to items in a stream.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def perform(self, obj):
        pass


class Predicate(object):
    """
    Used to apply a boolean test to items in a stream, to determine if it meets some set
    criteria.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def test(self, obj):
        pass


class Function(object):
    """
    Applies a function to items in a stream, producing an appropriate result.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def apply(self, input):
        pass




import abc
import multiprocessing
import Queue


class BufferedQueue(object):
    """
    Abstract implementation of a queue with a finite capacity. If the queue becomes full and
    further items are added, then those elements are dropped. In addition, if an item is
    requested from the queue and no items are present, then the system waits for an item to be
    added before returning.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def offer(self, item):
        """
        Inserts the specified element into the queue, but only if it is possible to do so
        immediately without violating capacity restrictions. Returns True if the item was added,
        and False if there is no space currently available.
        """
        pass

    @abc.abstractmethod
    def take(self):
        """
        Retrieves and removes the head of this queue, waiting if necessary until an element
        becomes available.
        """
        pass

    @abc.abstractmethod
    def remaining_capacity(self):
        """
        Returns the number of additional elements that this queue can hold (assuming no other
        constraints).
        """
        pass


class QueueBufferedQueue(BufferedQueue):
    """
    Implementation of a BufferedQueue using Python's Queue library as the backing data structure.
    """
    def __init__(self, capacity):
        self.max_size = capacity
        self.queue = multiprocessing.Queue(capacity)

    def offer(self, item):
        try:
            self.queue.put(item, False)
        except Queue.Full:
            return False

        return True

    def take(self):
        return self.queue.get(True)

    def remaining_capacity(self):
        return self.remaining_capacity() - self.queue.qsize()



import abc


class Stream(object):
    """
    Abstract implementation of a data stream.
    """

    __metaclass__ = abc.ABCMeta

    def for_each(self, operation, limit=0):
        """
        Applies the given Operation to each item in the stream. The Operation executes on the
        items in the stream in the order that they appear in the stream.

        If the limit is supplied, then processing of the stream will stop after that many items
        have been processed.
        """
        if limit != 0:
            count = 0
            while self.has_next():
                operation.perform(self.next())
                count += 1
                if count >= limit:
                    break
        else:
            while self.has_next():
                operation.perform(self.next())

    def filter(self, predicate):
        """
        Transforms the stream by only keeping items that match the supplied predicate.
        """
        return FilterStream(self, predicate)

    def map(self, function):
        """
        Transforms the stream by applying the supplied function to each item in the stream,
        thus creating a new stream.
        """
        return MapStream(self, function)

    @abc.abstractmethod
    def has_next(self):
        """
        Tests to see if there are any items left in the stream to consume.
        """
        pass

    @abc.abstractmethod
    def next(self):
        """
        Fetches the next item in the stream.
        """
        pass


class FilterStream(Stream):
    """
    A stream created by applying a filter (in the form of a Predicate) to another stream.
    """

    def __init__(self, source, predicate):
        self.source = source
        self.predicate = predicate
        self.obj = None

    def has_next(self):
        if self.obj is not None:
            return True

        while self.source.has_next() and self.obj is None:
            self.obj = self.source.next()
            if not self.predicate.test(self.obj):
                self.obj = None

        return self.obj is not None

    def next(self):
        if not self.has_next():
            raise Exception("Iteration has no more elements")

        to_return = self.obj
        self.obj = None

        return to_return


class MapStream(Stream):
    """
    A stream created by applying a Function to the elements in another stream.
    """

    def __init__(self, source, function):
        self.source = source
        self.function = function

    def has_next(self):
        return self.source.has_next()

    def next(self):
        return self.function.apply(self.source.next())


class BufferedStream(Stream):
    """
    Implementation of a Stream that uses a BufferedQueue as its internal buffer.

    This class is designed for use with live data sources that may produce data faster than it
    can be consumed, as the internal BufferedQueue will drop items that aren't consumed (i.e,
    removed from the queue) fast enough.
    """

    def __init__(self, buf):
        self.buf = buf
        self.connected = False

    def register(self, item):
        """
        Attempts to 'register' an item with the BufferedStream by offering it to the
        BufferedQueue. Returns True if the item was successfully published to the stream, or False
        if it wasn't.
        """
        return self.buf.offer(item)

    def connect(self):
        """
        Opens the streaming connection to the data source (makes has_next() return True)
        """
        self.connected = True

    def disconnect(self):
        """
        Closes the stream (by making has_next() return False)
        """
        self.connected = False

    def has_next(self):
        return self.connected

    def next(self):
        return self.buf.take()


from setuptools import setup

setup(name="spout",
      version="0.1.0",
      description="A simple framework that makes it easy to work with data streams in Python.",
      url="http://github.com/daviesjamie/spout",
      author="Jamie Davies",
      author_email="jamie@jamiedavies.me",
      license="MIT",
      packages=['spout'],
      zip_safe=False)
